/* implementation file for init., add, find, and clear  functions for
   sorted list data structure.  implemented as an AVL tree. */

#include <string.h>
#include <stdlib.h>

#include "align.h"
#include "stckallc.h"
#include "sortlist.h"
#include "slintrnl.h"

/* initialize sorted list */
void init_sort_list
  (
    /* pointer to sorted list to initialize */
    SORT_LIST *sl,
    /* sizeof elements to be stored in list */
    size_t e_size,
    /* pointer to element - element compare function.  the two parameters
       are pointers to elements.  this function should return some
       negative value if the key associated with the first element is
       less than the key associated with the second element.  it should
       return 0 if the key associated with the first element is equal to
       the key associated with the second element.  it should return some
       positive value if the key associated with the first element is
       greater than the key associated with the second element. */
    int (*e_e_compare)(const void *elem1, const void *elem2),
    /* pointer to key - element compare function.  the first paramter
       is a pointer to a key.  the second parameter is a pointer to
       an element.  this function should return some negative value
       if the key passed as a parameter is less than the key associated
       with the element parameter.  it should return 0 if the key
       parameter is equal to the key associated with the element
       parameter.  it should return some positive value if the key
       parameter is greater than the key associated with the element
       parameter. */
    int (*k_e_compare)(const void *key, const void *elem)
  )
  {
    sl->elem_size = e_size;
    init_alloc_stack(&(sl->node_alloc_stack),
                     N_UNITS_IN_NODE_PREFIX * sizeof(ALIGN_TYPE)
                     + e_size);
    sl->elem_elem_compare = e_e_compare;
    sl->key_elem_compare = k_e_compare;
    sl->tree = (void *) 0;

    return;
  }

/* balance an unbalanced tree.  returns non-zero if depth of
   tree reduced, otherwise 0. */
int balance_tree
  (
    /* pointer to pointer to subtree of sorted list to balance */
    void **t
  )
  {
    /* the three nodes to juggle */
    void *(node[3]);
    /* the top level branch that is longer */
    int top_level_branch;
    /* flag which indicates if the second level longer branch is the
       same as the top level branch (or both branches equal at
       2nd level) */
    int same_2nd_level;
    /* flag which indicates if the depth or the tree will be
       reduced */
    int reduce = 1;

    node[0] = *t;
    top_level_branch = NODE(node[0])->balance > 0 ? GREATER_BRANCH : 
						    LESS_BRANCH;

    node[1] = NODE(node[0])->branch[top_level_branch];

    if (NODE(node[1])->balance == 0)
      {
	same_2nd_level = 1;
	reduce = 0;
      }
    else
      {
	same_2nd_level = NODE(node[1])->balance > 0 ? GREATER_BRANCH : 
						      LESS_BRANCH;
	same_2nd_level = same_2nd_level == top_level_branch;
      }

    if (same_2nd_level)
      node[2] = NODE(node[1])->branch[top_level_branch];
    else
      {
	node[2] = node[1];
	node[1] = NODE(node[1])->branch[OTHER(top_level_branch)];
      }

    /* move appropriate subtrees */
    NODE(node[0])->branch[top_level_branch] =
      NODE(node[1])->branch[OTHER(top_level_branch)];
    if (!same_2nd_level)
      NODE(node[2])->branch[OTHER(top_level_branch)] =
	NODE(node[1])->branch[top_level_branch];

    /* make node 1 new root node */
    NODE(node[1])->branch[OTHER(top_level_branch)] = node[0];
    NODE(node[1])->branch[top_level_branch] = node[2];
    *t = node[1];

    /* set balance factors */
    if (!reduce)
      {
	int bal = top_level_branch == GREATER_BRANCH ? 1 : -1;
	NODE(node[1])->balance = -bal;
	NODE(node[0])->balance = bal;
	/* balance factor for node 2 stays the same */
      }
    else
      {
	if (same_2nd_level)
	  NODE(node[0])->balance = 0;
	  /* balance factor for node 2 stays the same */
	else
	  {
	    char bal[3];

	    bal[1] = NODE(node[1])->balance;
	    if (top_level_branch == LESS_BRANCH)
	      bal[1] = -bal[1];
	    bal[0] = 0;
	    bal[2] = 0;
	    if (bal[1] == -1)
	      bal[2] = 1;
	    else if (bal[1] == 1)
	      bal[0] = -1;
	    if (top_level_branch == LESS_BRANCH)
	      {
		bal[0] = -bal[0];
		bal[2] = -bal[2];
	      }
	    NODE(node[0])->balance = bal[0];
	    NODE(node[2])->balance = bal[2];
	  }
	NODE(node[1])->balance = 0;
      }

    return(reduce);
  }

/* add an element to a sorted list.  memory will be allocated
   for the element, and it will be copied to this memory using
   memcpy().  the function returns SL_SUCCESS if successful.  if the
   allocation of memory for the copy of the element fails, the
   function returns SL_MEM_ALLOC_FAIL.  if an element with the
   same key as the element to be added is already in the list,
   SL_DUP_KEY is returned. */
SL_RESULT add_sort_list
  (
    /* sorted list to add element to */
    SORT_LIST *sl,
    /* element to add to list */
    const void *elem
  )
  {
    /* pointer to pointer to smallest unbalanced subtree
       that will contain the element to be inserted */
    void **unbal_subtree = (void **) 0;
    /* pointer to pointer to current node */
    void **p_node;
    /* new node to be inserted */
    void *new_node;

    void *node;
    int which_branch, i;

    /* create a node for the element */
    new_node = get_alloc_stack(&sl->node_alloc_stack);
    if (!new_node)
      return(SL_MEM_ALLOC_FAIL);
    memcpy(ELEM_IN_NODE(new_node), elem, sl->elem_size);
    NODE(new_node)->balance = 0;
    NODE(new_node)->branch[LESS_BRANCH] = (void *) 0;
    NODE(new_node)->branch[GREATER_BRANCH] = (void *) 0;

    if (!sl->tree)
      {
	sl->tree = new_node;
	return(SL_SUCCESS);
      }

    p_node = &(sl->tree);
    node = sl->tree;
    for ( ; ; )
      {
	i = (*(sl->elem_elem_compare))(elem, ELEM_IN_NODE(node));

	if (i == 0)
	  /* duplicate key */
	  {
	    /* free the node that was allocated */
	    free_alloc_stack(&(sl->node_alloc_stack));
	    return(SL_DUP_KEY);
	  }

	if (NODE(node)->balance)
	  unbal_subtree = p_node;

	which_branch = i > 0 ? GREATER_BRANCH : LESS_BRANCH;

	if (!NODE(node)->branch[which_branch])
	  /* place to insert */
	  {
	    NODE(node)->branch[which_branch] = new_node;
	    break;
	  }

	p_node = &(NODE(node)->branch[which_branch]);
	node = *p_node;
      }

    /* adjust balance factors */
    node = unbal_subtree ? *unbal_subtree : sl->tree;
    for ( ; ; )
      {
	i = (*(sl->elem_elem_compare))(elem, ELEM_IN_NODE(node));

	if (i == 0)
	  break;

	NODE(node)->balance += i > 0 ? 1 : -1;

	node = NODE(node)->branch[i > 0 ? GREATER_BRANCH : LESS_BRANCH];
      }

    /* rebalance subtree if necessary */
    if (unbal_subtree)
      {
	node = *unbal_subtree;
	i = NODE(node)->balance;
	if ((i == 2) || (i == -2))
	  /* depth of tree will always be reduced when adding */
	  (void) balance_tree(unbal_subtree);
      }

    return(SL_SUCCESS);
  }

/* find the element which matches the given key based on the given
   match type.  returns pointer to element if match is found.  returns
   null if no match found.  the calling code must not change the
   key value associated with the element. */
void *find_sort_list
  (
    /* sorted list to search */
    const SORT_LIST *sl,
    /* code for type of match desired */
    unsigned int match_type,
    /* key to match.  this parameter is a don't care if match_type
       is MATCH_LEAST or MATCH_GREATEST. */
    const void *key
  )
  {
    void *node;
    int which_branch;

    if (match_type & (MATCH_LEAST | MATCH_GREATEST))
      /* find min. or max. element */
      {
	which_branch = match_type == MATCH_LEAST ? LESS_BRANCH :
						   GREATER_BRANCH;
        node = sl->tree;

	if (!node)
	  /* tree is empty */
	  return((void *) 0);

	for ( ; ; )
	  {
	    if (!NODE(node)->branch[which_branch])
	      /* the element in this node is the min. or max. */
	      break;
	    /* continue searching */
	    node = NODE(node)->branch[which_branch];
	  }
      }
    else
      {
	int compare_result;
	void *p = sl->tree;

	node = (void *) 0;

	while (p) 
	  {
	    /* compare with current element */
	    compare_result =
              (*sl->key_elem_compare)(key, ELEM_IN_NODE(p));

	    if ((match_type & MATCH_EQUAL) && (compare_result == 0))
	      /* success */
	      {
	        node = p;
	        break;
	      }

	    if (((match_type & MATCH_GREATER) && (compare_result < 0)) ||
	        ((match_type & MATCH_LESS) && (compare_result > 0)))
	      /* best match so far found - continue looking for better
		 matches */
	      node = p;

	    if (compare_result == 0)
	      /* MATCH_EQUAL found when it wasn't desired.  need
		 to set a dummy compare result to continue searching
		 down the right branch */
	      {
		if (match_type == MATCH_GREATER)
		  compare_result = 1;
		else
		  compare_result = -1;
	      }

	    which_branch = 
              compare_result > 0 ? GREATER_BRANCH : LESS_BRANCH;

	    /* proceed to next subtree */
	    p = NODE(p)->branch[which_branch];

	  } /* while (p) */

	if (!node)
	  return((void *) 0);

      } /* else */

    return(ELEM_IN_NODE(node));
  }

/* delete all the elements in a sorted list */
void clear_sort_list
  (
    /* sorted list to search */
    SORT_LIST *sl
  )
  {
    /* free all nodes */
    clear_alloc_stack(&(sl->node_alloc_stack));

    sl->tree = (void *) 0;
  }
